# Software License Agreement (BSD License)
#
# Copyright (c) 2013, Open Source Robotics Foundation, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Open Source Robotics Foundation, Inc. nor
#    the names of its contributors may be used to endorse or promote
#    products derived from this software without specific prior
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import re

from .vcs import Git


class RepositorySpecification(object):
    # Match groups are server and path on server.
    VCS_REGEX = re.compile('(?:https?:\/\/|ssh:\/\/|git:\/\/|git@)((?:[a-fA-F0-9]{40}@)?[\w.-]+)[:/]([\w/-]*)(?:\.git)?$')

    def __init__(self, name, data):
        self.name = name
        self.type = data.get('type', 'git')
        assert 'url' in data and data['url'], "Repository '%s' lacks required URL information" % name
        self.url = data['url']
        self.version = data.get('version', None)
        self._remote_refs = None

        # for backward compatibility only
        self.status = None
        self.status_description = None

    def get_data(self):
        return self._get_data()

    def get_url_parts(self):
        """ Returns a tuple for the server and path, eg ('github.com', 'ros/catkin') """
        match = self.VCS_REGEX.match(self.url)
        if not match:
            raise RuntimeError('VCS url "%s" does not match expected format.' % self.url)
        return match.groups()

    def has_remote_tag(self, tag):
        return tag in self.remote_tags

    @property
    def remote_refs(self):
        if not self._remote_refs:
            result = Git().command('ls-remote', self.url)
            if result['returncode'] != 0:
                raise RuntimeError('Could not git ls-remote repository "%s"' % self.url)
            self._remote_refs = {}
            for row in result['output'].strip().splitlines():
                sha, name = row.split('\t')
                self._remote_refs[name] = sha
        return self._remote_refs

    @property
    def remote_tags(self):
        result = {}
        for name, sha in self.remote_refs.items():
            if name.startswith('refs/tags/'):
                result[name.split('refs/tags/')[1]] = sha
        return result

    def _get_data(self, skip_git_type=False):
        data = {}
        if self.type != 'git' or not skip_git_type:
            data['type'] = str(self.type)
        data['url'] = str(self.url)
        if self.version is not None:
            data['version'] = self.version
        return data



# Software License Agreement (BSD License)
#
# Copyright (c) 2018, Open Source Robotics Foundation, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Open Source Robotics Foundation, Inc. nor
#    the names of its contributors may be used to endorse or promote
#    products derived from this software without specific prior
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


class SourceRepositoryCache(object):
    """
    This class represents a cache of the package XML strings for all packages in a single
    repo at a particular moment in time. A dictionary of many of these (one for each repo)
    keyed to the repo name represents the totality of the source package xml cache.
    """

    def __init__(self, data):
        assert data
        self._ref = data['_ref']
        self._package_names = set([name for name in data.keys() if name != '_ref'])
        self._data = data

    def get_data(self):
        """
        Return the bare data dict, suitable for serializing into yaml.
        """
        return self._data

    @classmethod
    def from_ref(cls, ref):
        """
        Create a new empty cache instance from just the version control reference hash.
        """
        return cls({'_ref': ref})

    def add(self, package_name, package_path, package_xml_string):
        """
        Add a package to the cache.
        """
        self._data[package_name] = (package_path, package_xml_string)
        self._package_names.add(package_name)

    def __iter__(self):
        """
        Iterate the list of package names in this cached repo. Returns an iterator of str.
        """
        return iter(self._package_names)

    def __getitem__(self, package_name):
        """
        Access the cached information about a specific package. Returns a (str, str) of
        path to package relative to repo root, and string of package xml.
        """ 
        if package_name not in self._package_names:
            raise KeyError("Package '%s' not present in SourceRepositoryCache." % package_name)
        return self._data[package_name]

    def items(self):
        """
        Generator of (str, str, str) containing the package name, path relative
        to repo root, and package xml string.
        """
        for package_name in self._package_names:
            package_path, package_xml_string = self._data[package_name]
            yield package_name, package_path, package_xml_string

    def __len__(self):
        """
        Returns the number of packages in this repo.
        """
        return len(self._package_names)

    def keys(self):
        """
        Return the list of packages in the repo cache.
        """
        return self._package_names

    def ref(self):
        """
        Return the version control hash.
        """
        return self._ref



# Software License Agreement (BSD License)
#
# Copyright (c) 2013, Open Source Robotics Foundation, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Open Source Robotics Foundation, Inc. nor
#    the names of its contributors may be used to endorse or promote
#    products derived from this software without specific prior
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

try:
    from urllib.request import urlopen, Request
    from urllib.error import URLError
except ImportError:
    from urllib2 import urlopen, Request
    from urllib2 import URLError

import base64
import json
import os

from catkin_pkg.package import parse_package_string

from rosdistro.source_repository_cache import SourceRepositoryCache
from rosdistro import logger

GITHUB_USER = os.getenv('GITHUB_USER', None)
GITHUB_PASSWORD = os.getenv('GITHUB_PASSWORD', None)

def _get_url_contents(url):
    return urlopen(url).read().decode('utf-8')

def github_manifest_provider(_dist_name, repo, pkg_name):
    assert repo.version
    server, path = repo.get_url_parts()
    if not server.endswith('github.com'):
        logger.debug('Skip non-github url "%s"' % repo.url)
        raise RuntimeError('can not handle non github urls')

    release_tag = repo.get_release_tag(pkg_name)

    if not repo.has_remote_tag(release_tag):
        raise RuntimeError('specified tag "%s" is not a git tag' % release_tag)

    url = 'https://raw.githubusercontent.com/%s/%s/package.xml' % (path, release_tag)
    try:
        logger.debug('Load package.xml file from url "%s"' % url)
        return _get_url_contents(url)
    except URLError as e:
        logger.debug('- failed (%s), trying "%s"' % (e, url))
        raise RuntimeError()


def github_source_manifest_provider(repo):
    server, path = repo.get_url_parts()
    if not server.endswith('github.com'):
        logger.debug('Skip non-github url "%s"' % repo.url)
        raise RuntimeError('can not handle non github urls')

    tree_url = 'https://api.github.com/repos/%s/git/trees/%s?recursive=1' % (path, repo.version)
    req = Request(tree_url)
    if GITHUB_USER and GITHUB_PASSWORD:
        logger.debug('- using http basic auth from supplied environment variables.')
        authheader = 'Basic %s' % base64.b64encode('%s:%s' % (GITHUB_USER, GITHUB_PASSWORD))
        req.add_header('Authorization', authheader)
    try:
        tree_json = json.loads(_get_url_contents(req))
        logger.debug('- load repo tree from %s' % tree_url)
    except URLError as e:
        raise RuntimeError('Unable to fetch JSON tree from %s: %s' % (tree_url, e))

    if tree_json['truncated']:
        raise RuntimeError('JSON tree is truncated, must perform full clone.')

    package_xml_paths = set()
    for obj in tree_json['tree']:
        if obj['path'].split('/')[-1] == 'package.xml':
            package_xml_paths.add(os.path.dirname(obj['path']))

    # Filter out ones that are inside other packages (eg, part of tests)
    def package_xml_in_parent(path):
        if path == '':
            return True
        parent = path
        while True:
            parent = os.path.dirname(parent)
            if parent in package_xml_paths:
                return False
            if parent == '':
                return True
    package_xml_paths = list(filter(package_xml_in_parent, package_xml_paths))

    cache = SourceRepositoryCache.from_ref(tree_json['sha'])
    for package_xml_path in package_xml_paths:
        url = 'https://raw.githubusercontent.com/%s/%s/%s' % \
            (path, cache.ref(), package_xml_path + '/package.xml' if package_xml_path else 'package.xml')
        logger.debug('- load package.xml from %s' % url)
        package_xml = _get_url_contents(url)
        name = parse_package_string(package_xml).name
        cache.add(name, package_xml_path, package_xml)

    return cache



import os

from rosdistro import get_index, get_source_file
from rosdistro.loader import load_url
from rosdistro.source_file import SourceFile

import yaml

FILES_DIR = os.path.normpath(os.path.join(os.path.dirname(os.path.abspath(__file__)), 'files'))


def test_source_file():
    url = 'file://' + FILES_DIR + '/foo/distribution.yaml'
    yaml_str = load_url(url)
    data = yaml.load(yaml_str)
    src_file = SourceFile('foo', data)
    _validate_src_file(src_file)


def test_get_source_file():
    url = 'file://' + FILES_DIR + '/index_v2.yaml'
    i = get_index(url)
    src_file = get_source_file(i, 'foo')
    _validate_src_file(src_file)


def _validate_src_file(src_file):
    assert(set(['bar_repo', 'baz-repo']) == set(src_file.repositories.keys()))

    repo = src_file.repositories['bar_repo']
    assert(repo.type == 'git')
    assert(repo.url == 'https://github.com/example-test/bar_repo.git')
    assert(repo.version == 'master')

    repo = src_file.repositories['baz-repo']
    assert(repo.type == 'hg')
    assert(repo.url == 'https://bitbucket.org/baz-test/baz-repo')
    assert(repo.version == 'default')



from __future__ import print_function

import sys

_quiet = False


def quiet(state=True):
    global _quiet
    _quiet = state


def _print_func(msg, end, file):
    global _quiet
    if not _quiet:
        print(msg, end=end, file=file)


def override_print(print_func=_print_func):
    global _print
    _print = print_func


_print = _print_func


def info(msg, end=None, file=None):
    global _print
    _print(msg, end=end, file=file)


def warning(msg, end=None, file=None):
    global _print
    _print(msg, end=end, file=file)


def error(msg, end=None, file=None):
    global _print
    file = sys.stderr if file is None else file
    _print(msg, end=end, file=file)



# Software License Agreement (BSD License)
#
# Copyright (c) 2013, Open Source Robotics Foundation, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Open Source Robotics Foundation, Inc. nor
#    the names of its contributors may be used to endorse or promote
#    products derived from this software without specific prior
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from .package import Package
from .release_repository_specification import ReleaseRepositorySpecification


class ReleaseFile(object):

    _type = 'distribution'

    def __init__(self, name, data):
        self.name = name

        assert 'type' in data and data['type'] != 'release', "Unable to handle 'release' format anymore, please update your 'release' file to the latest specification"
        assert 'type' in data, "Expected file type is '%s'" % ReleaseFile._type
        assert data['type'] == ReleaseFile._type, "Expected file type is '%s', not '%s'" % (ReleaseFile._type, data['type'])

        assert 'version' in data, "Release file for '%s' lacks required version information" % self.name
        assert int(data['version']) == 1, "Unable to handle '%s' format version '%d', please update rosdistro (e.g. on Ubuntu/Debian use: sudo apt-get update && sudo apt-get install --only-upgrade python-rosdistro)" % (ReleaseFile._type, int(data['version']))
        self.version = int(data['version'])

        self.repositories = {}
        self.packages = {}
        if 'repositories' in data:
            for repo_name in sorted(data['repositories'].keys()):
                if 'release' not in data['repositories'][repo_name]:
                    continue
                repo_status_data = data['repositories'][repo_name]
                repo_data = repo_status_data['release']
                try:
                    repo = ReleaseRepositorySpecification(repo_name, repo_data)
                    # for backward compatibility only
                    repo.status = repo_status_data.get('status', None)
                    repo.status_description = repo_status_data.get('status_description', None)
                except AssertionError as e:
                    e.args = [("Release file '%s': %s" % (self.name, a) if i == 0 else a) for i, a in enumerate(e.args)]
                    raise e
                self.repositories[repo_name] = repo

                if repo.package_names:
                    for pkg_name in repo.package_names:
                        assert pkg_name not in self.packages, "Duplicate package name '%s' exists in repository '%s' as well as in repository '%s'" % (pkg_name, repo_name, self.packages[pkg_name].repository_name)
                        self._add_package(pkg_name, repo, unary_repo=len(repo.package_names) == 1 and pkg_name == repo_name)
                else:
                    # no package means a single package in the root of the repository
                    self._add_package(repo_name, repo, {'subfolder': '.'}, True)

        self.platforms = {}
        if 'release_platforms' in data:
            for os_name in data['release_platforms'].keys():
                self.platforms[os_name] = []
                for os_code_name in data['release_platforms'][os_name]:
                    assert os_code_name not in self.platforms[os_name], "Distribution '%s' specifies the os_code_name '%s' multiple times for the os_name '%s'" % (self.name, os_code_name, os_name)
                    self.platforms[os_name].append(os_code_name)

    def _add_package(self, pkg_name, repo, unary_repo):
        assert pkg_name not in self.packages
        self.packages[pkg_name] = Package(pkg_name, repo.name)



# Software License Agreement (BSD License)
#
# Copyright (c) 2013, Open Source Robotics Foundation, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Open Source Robotics Foundation, Inc. nor
#    the names of its contributors may be used to endorse or promote
#    products derived from this software without specific prior
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from .repository_specification import RepositorySpecification


class ReleaseRepositorySpecification(RepositorySpecification):

    def __init__(self, name, data):
        super(ReleaseRepositorySpecification, self).__init__(name, data)
        assert self.type == 'git'

        self.tags = {}
        if self.version is not None:
            assert 'tags' in data
            assert 'release' in data['tags']
        if 'tags' in data:
            for tag_type in data['tags']:
                tag_data = data['tags'][tag_type]
                self.tags[tag_type] = str(tag_data)

        self.package_names = []
        if 'packages' in data and data['packages']:
            self.package_names = sorted(data['packages'])
        else:
            # no packages means a single package
            self.package_names = [self.name]

        # for backward compatibility only
        self.release_repository = self

    def get_release_tag(self, pkg_name):
        data = {
            'package': pkg_name
        }
        if self.version:
            data['version'] = self.version
            data['upstream_version'] = self.version.split('-')[0]
        release_tag = self.tags['release']
        for k, v in data.items():
            release_tag = release_tag.replace('{%s}' % k, v)
        return release_tag

    def get_data(self):
        data = self._get_data(skip_git_type=True)
        if self.tags:
            data['tags'] = {}
            for tag in self.tags:
                data['tags'][tag] = str(self.tags[tag])
        if len(self.package_names) > 1 or (len(self.package_names) == 1 and self.package_names[0] != self.name):
            data['packages'] = sorted(self.package_names)
        return data



# Software License Agreement (BSD License)
#
# Copyright (c) 2016, Clearpath Robotics
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Open Source Robotics Foundation, Inc. nor
#    the names of its contributors may be used to endorse or promote
#    products derived from this software without specific prior
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import print_function

import subprocess
import sys
import threading
import time

try:
    import queue
except ImportError:
    import Queue as queue

CONCURRENT_DEFAULT = 16


def freeze_distribution_sources(dist, release_version=False, release_tag=False,
                                concurrent_ops=CONCURRENT_DEFAULT, quiet=False):
    # Populate this queue with tuples of repositories instances to be updated,
    # so that this work can be spread across multiple threads.
    work_queue = queue.Queue()
    for repo_name, repo in dist.repositories.items():
        # Only manipulate distribution entries with a source repo listed.
        if repo.source_repository:
            # Decide which git ref string we'll be using as the replacement match.
            if repo.release_repository and (release_version or release_tag):
                version = repo.release_repository.version.split('-')[0]
            else:
                version = repo.source_repository.version
            work_queue.put((repo.source_repository, version, release_tag))

    total_items = work_queue.qsize()

    for i in range(concurrent_ops):
        threading.Thread(target=_worker, args=[work_queue]).start()

    # Wait until the threads have done all the work and exited.
    while not work_queue.empty():
        time.sleep(0.1)
        if not quiet:
            sys.stdout.write("Updating source repo versions (%d/%d)   \r" %
                             (total_items - work_queue.qsize(), total_items))
            sys.stdout.flush()
    work_queue.join()

    # Clear past the updating line.
    if not quiet:
        print("")


# Get the repo commit information
def _get_repo_info(url, retry=2, retry_period=1):
    cmd = ['git', 'ls-remote', url]
    try:
        return subprocess.check_output(cmd).decode().splitlines()
    except subprocess.CalledProcessError as err:
        if not retry:
            raise
        print('  Non-zero return code for: %s, retrying in %f seconds' %
              (' '.join(cmd), retry_period), file=sys.stderr)
        # brief delay incase its an intermittent issue with infrastructure
        time.sleep(retry_period)
        return _get_repo_info(url, retry=retry - 1, retry_period=retry_period * 2)


def _worker(work_queue):
    while True:
        try:
            source_repo, freeze_version, freeze_to_tag = work_queue.get(block=False)
            ls_remote_lines = _get_repo_info(source_repo.url)
            for line in ls_remote_lines:
                hash, ref = line.split('\t', 1)
                if freeze_to_tag and ref == 'refs/tags/%s' % freeze_version:
                    source_repo.version = ref.split('refs/tags/')[1]
                    break
                elif ref in ('refs/heads/%s' % freeze_version, 'refs/tags/%s' % freeze_version):
                    source_repo.version = hash
                    break

            if not isinstance(source_repo.version, str):
                # On Python 2, explicitly encode back to string. Unnecessary on Python 3, because
                # we're already a unicode string since subprocess.check_output returned a bytes.
                source_repo.version = source_repo.version.encode()

            work_queue.task_done()

        except subprocess.CalledProcessError as e:
            print("No information could be retrieved for repo %s with error: %s" %
                  (source_repo.url, e), file=sys.stderr)
            work_queue.task_done()

        except queue.Empty:
            break



import os

from rosdistro import get_distribution_file, get_distribution_files, get_index
from rosdistro.distribution_file import DistributionFile
from rosdistro.loader import load_url

import yaml

FILES_DIR = os.path.normpath(os.path.join(os.path.dirname(os.path.abspath(__file__)), 'files'))


def test_distribution_file():
    url = 'file://' + FILES_DIR + '/foo/distribution.yaml'
    yaml_str = load_url(url)
    data = yaml.load(yaml_str)
    dist_file = DistributionFile('foo', data)
    _validate_dist_file(dist_file)


def test_get_distribution_file():
    url = 'file://' + FILES_DIR + '/index_v2.yaml'
    i = get_index(url)
    dist_file = get_distribution_file(i, 'foo')
    _validate_dist_file(dist_file)

    dist_files = get_distribution_files(i, 'foo')
    assert len(dist_files) == 1


def _validate_dist_file(dist_file):
    assert('bar_repo' in dist_file.repositories)
    repo = dist_file.repositories['bar_repo']
    assert repo.release_repository.package_names == ['bar_repo']
    assert 'bar_repo' in dist_file.release_packages

    assert'baz-repo' in dist_file.repositories
    repo = dist_file.repositories['baz-repo']
    assert set(repo.release_repository.package_names) == set(['baz_pkg1', 'baz_pkg2'])
    assert 'baz_pkg1' in dist_file.release_packages
    assert 'baz_pkg2' in dist_file.release_packages



# Software License Agreement (BSD License)
#
# Copyright (c) 2013, Open Source Robotics Foundation, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Open Source Robotics Foundation, Inc. nor
#    the names of its contributors may be used to endorse or promote
#    products derived from this software without specific prior
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from .manifest_provider.bitbucket import bitbucket_manifest_provider
from .manifest_provider.git import git_manifest_provider, git_source_manifest_provider
from .manifest_provider.github import github_manifest_provider, github_source_manifest_provider
from .package import Package


class Distribution(object):

    default_manifest_providers = [github_manifest_provider, bitbucket_manifest_provider, git_manifest_provider]
    default_source_manifest_providers = [github_source_manifest_provider, git_source_manifest_provider]

    def __init__(self, distribution_file, manifest_providers=None, source_manifest_providers=None):
        self._distribution_file = distribution_file

        # Use default
        self._manifest_providers = Distribution.default_manifest_providers
        self._source_manifest_providers = Distribution.default_source_manifest_providers

        # Override default if given
        if manifest_providers is not None:
            self._manifest_providers = manifest_providers
        if source_manifest_providers is not None:
            self._source_manifest_providers = source_manifest_providers

        self._release_package_xmls = {}
        self._source_repo_package_xmls = {}

    def __getattr__(self, name):
        return getattr(self._distribution_file, name)

    def get_release_package_xml(self, pkg_name):
        if pkg_name not in self._release_package_xmls:
            pkg = self._distribution_file.release_packages[pkg_name]
            repo_name = pkg.repository_name
            repo = self._distribution_file.repositories[repo_name]
            if repo.release_repository is None:
                return None
            repo = repo.release_repository
            if repo.version is None:
                return None
            package_xml = None
            for mp in self._manifest_providers:
                package_xml = mp(self._distribution_file.name, repo, pkg_name)
                if package_xml is not None:
                    break
            self._release_package_xmls[pkg_name] = package_xml
        return self._release_package_xmls[pkg_name]

    def get_source_package_xml(self, pkg_name):
        repo_name = self._distribution_file.source_packages[pkg_name].repository_name
        repo_cache = self.get_source_repo_package_xmls(repo_name)
        if repo_cache:
            return repo_cache[pkg_name][1]
        else:
            return None

    def get_source_repo_package_xmls(self, repo_name):
        if repo_name in self._source_repo_package_xmls:
            return self._source_repo_package_xmls[repo_name]
        else:
            for mp in self._source_manifest_providers:
                repo_cache = mp(self.repositories[repo_name].source_repository)
                if repo_cache is not None:
                    # Update map of package XMLs, and also list of known package names.
                    self._source_repo_package_xmls[repo_name] = repo_cache
                    for pkg_name in repo_cache:
                        if pkg_name[0] != '_':
                            self._distribution_file.source_packages[pkg_name] = Package(pkg_name, repo_name)
                    return self._source_repo_package_xmls[repo_name]
        return None
